
import numpy as np
from tensorflow import tensordot, concat, sin, cos
from tensorflow.python import keras as k
from tensorflow.python.util.tf_export import keras_export
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.keras.engine.input_spec import InputSpec
from tensorflow.python.keras.initializers import Constant as default_constant_initializer


@keras_export('keras.layers.SciActivationLayer')
class FourierLayer(k.layers.Layer):

    def __init__(self, w=10, **kwargs):
        super(FourierLayer, self).__init__(**kwargs)
        if isinstance(w, int):
            w = 2*np.pi/np.random.rand(w)
        else:
            assert isinstance(w, np.ndarray)
        self.frequencies = w.flatten()
        self.num_features = w.size

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        param_shape = list(input_shape[1:])
        assert param_shape == [1]
        w = self.frequencies.reshape(param_shape + [self.num_features])
        self.frequencies = k.backend.cast_to_floatx(w)
        self.built = True

    def call(self, inputs):
        xs = tensordot(inputs, self.frequencies, axes=1)
        return concat([sin(xs), cos(xs)], axis=-1)

    def get_config(self):
        config = {
            'frequencies': initializers.serialize(self.frequencies),
        }
        base_config = super(FourierLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        return [self.num_features]




""" Built-in utilities to process inputs.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import numpy as np
from numpy import pi

import tensorflow as tf
from tensorflow.python import keras as k
from tensorflow.python.keras import backend as K

# interface for some keras features to be acessible across sciann.
from tensorflow.python.keras.backend import is_keras_tensor as is_tensor
from tensorflow.python.keras.backend import floatx
from tensorflow.python.keras.backend import set_floatx
from tensorflow.python.keras.utils.vis_utils import model_to_dot
from tensorflow.python.keras.utils.vis_utils import plot_model
from tensorflow.python.keras.initializers import RandomUniform as default_bias_initializer
from tensorflow.python.keras.initializers import GlorotNormal as default_kernel_initializer
from tensorflow.python.keras.initializers import Constant as default_constant_initializer
from tensorflow.python.keras.initializers import VarianceScaling
from tensorflow.python.keras.regularizers import l1_l2

from .initializers import SciKernelInitializer as KInitializer
from .initializers import SciBiasInitializer as BInitializer
from .activations import get_activation, SciActivation, SciActivationLayer, SciRowdyActivationLayer
from .validations import is_functional

from pybtex.database.input import bibtex


_DEFAULT_LOG_PATH = ""
_BIBLIOGRAPHY = None
_BIBLIOGRAPHY_TO_OUTPUT = []


def _is_tf_1():
    return tf.__version__.startswith('1.')


def set_random_seed(val=1234):
    """ Set random seed for reproducibility.

    # Arguments
        val: A seed value..

    """
    np.random.seed(val)
    if _is_tf_1():
        tf.set_random_seed(val)
    else:
        tf.random.set_seed(val)


def reset_session():
    """ Clear keras and tensorflow sessions.
    """
    if _is_tf_1():
        K.clear_session()
    else:
        tf.keras.backend.clear_session()


clear_session = reset_session


def set_default_log_path(path):
    global _DEFAULT_LOG_PATH
    _DEFAULT_LOG_PATH = path


def get_default_log_path():
    return _DEFAULT_LOG_PATH


def initialize_bib(bib_file):
    global _BIBLIOGRAPHY
    global _BIBLIOGRAPHY_TO_OUTPUT
    _BIBLIOGRAPHY = bibtex.Parser().parse_file(bib_file)
    _BIBLIOGRAPHY_TO_OUTPUT.append(_BIBLIOGRAPHY.entries['haghighat2021sciann'])
    _BIBLIOGRAPHY_TO_OUTPUT.append(_BIBLIOGRAPHY.entries['raissi2019physics'])


def append_to_bib(bib_entery):
    global _BIBLIOGRAPHY
    global _BIBLIOGRAPHY_TO_OUTPUT
    for bib_entery_i in to_list(bib_entery):
        bib = _BIBLIOGRAPHY.entries[bib_entery_i]
        if bib not in _BIBLIOGRAPHY_TO_OUTPUT:
            _BIBLIOGRAPHY_TO_OUTPUT.append(bib)


def get_bibliography(format='bibtex', file_name=None):
    """Returns the bibliography based on the feastures you used in your model.

    # Argument
        format: 'bibtex', 'bibtextml', 'yaml', ...
            check `pybtex` documentation for other options.
            default: 'bibtex'
        file_name: path to a file.
            default: None. This results in printing the bib file in the outputs.
    """
    global _BIBLIOGRAPHY_TO_OUTPUT
    bib = ""
    for b in _BIBLIOGRAPHY_TO_OUTPUT:
        bib += str(b.to_string(format)) + '\n'

    if file_name is None:
        print(bib)
    else:
        with open(file_name,'w') as f:
            f.write(bib)
            f.close()


def get_log_path(path=None, prefix=None):
    file_path = _DEFAULT_LOG_PATH if path is None else path
    if not os.path.isdir(file_path):
        os.makedirs(file_path)
    # add the prefix.
    if prefix is None:
        return file_path
    else:
        return os.path.join(file_path, prefix)


def is_same_tensor(x, y):
    if len(to_list(x)) != len(to_list(y)):
        return False
    else:
        res = []
        for xi, yi in zip(to_list(x), to_list(y)):
            res.append(xi.name == yi.name)
        return all(res)


def unique_tensors(Xs):
    if len(Xs) > 1:
        ux, uids = np.unique([x.name for x in Xs], return_index=True)
        uids = sorted(uids)
        return [Xs[i] for i in uids]
    else:
        return Xs


def default_regularizer(*args, **kwargs):
    l1, l2 = 0.0, 0.0
    if (len(args) == 0 and len(kwargs) == 0) or args[0] is None:
        return None
    elif len(args) == 1:
        if isinstance(args[0], (float, int)):
            l1 = 0.0
            l2 = args[0]
        elif isinstance(args[0], list):
            l1 = args[0][0]
            l2 = args[0][1]
        elif isinstance(args[0], dict):
            l1 = 0.0 if 'l1' not in args[0] else args[0]['l1']
            l2 = 0.0 if 'l2' not in args[0] else args[0]['l2']
    elif len(args) == 2:
        l1 = args[0]
        l2 = args[1]
    elif len(kwargs) > 0:
        l1 = 0.0 if 'l1' not in kwargs else kwargs['l1']
        l2 = 0.0 if 'l2' not in kwargs else kwargs['l2']
    else:
        raise ValueError('Unrecognized entry - input regularization values for l1 and l2.')
    # print("regularization is used with l1={} and l2={}".format(l1, l2))
    return l1_l2(l1=l1, l2=l2)


def default_weight_initializer(actf='linear', distribution='uniform', mode='fan_in', scale=None):
    inz = []
    for i, af in enumerate(to_list(actf)):
        if distribution in ('uniform', 'normal'):
            tp = VarianceScaling(
                scale=eval_default_scale_factor(af, i) if scale is None else scale,
                mode=mode, distribution=distribution
            )
        elif distribution in ('constant',):
            tp = default_constant_initializer(0.0 if scale is None else scale)
        else:
            raise ValueError('Undefined distribution: pick from ("uniform", "normal", "constant").')
        inz.append(tp)
    return inz


def eval_default_scale_factor(actf, lay):
    if actf in ('linear', 'relu'):
        return 2.0
    elif actf in ('tanh', 'sigmoid'):
        return 1.0 if lay > 0 else 1.0
    elif actf in ('sin', 'cos'):
        return 2.0 if lay > 0 else 2.0 #*30.0
    else:
        return 1.0


def prepare_default_activations_and_initializers(actfs, seed=None):
    activations = []
    bias_initializer = []
    kernel_initializer = []
    for lay, actf in enumerate(to_list(actfs)):
        # initializers.
        bias_initializer.append(BInitializer(lay, seed))
        kernel_initializer.append(KInitializer(lay, seed))
        w = kernel_initializer[-1].w0
        # support rowdy net.
        layer_actfs = []
        for sig in to_list(actf):
            if isinstance(sig, str):
                lay_actf = sig.lower().split('-')
                f = get_activation(lay_actf[-1])
            elif callable(sig):
                lay_actf = sig.__name__.lower().split('-')
                f = sig
            else:
                raise TypeError('expected a string for actf: {}'.format(sig))
            if len(lay_actf) == 2:
                append_to_bib("jagtap2020locally")
                layer_actfs.append(SciActivationLayer(w, f, lay_actf[0]))
            else:
                layer_actfs.append(SciActivation(w, f))
        if len(layer_actfs) == 1:
            sig_layer = layer_actfs[0]
        else:
            append_to_bib("jagtap2021deep")
            sig_layer = SciRowdyActivationLayer(layer_actfs)
        activations.append(sig_layer)

    return activations, bias_initializer, kernel_initializer


def unpack_singleton(x):
    """Gets the first element if the iterable has only one value.

    Otherwise return the iterable.

    # Argument
        x: A list or tuple.

    # Returns
        The same iterable or the first element.
    """
    if len(x) == 1:
        return x[0]
    return x


def to_list(x, allow_tuple=False):
    """Normalizes a list/tensor into a list.

    If a tensor is passed, we return
    a list of size 1 containing the tensor.

    # Arguments
        x: target object to be normalized.
        allow_tuple: If False and x is a tuple,
            it will be converted into a list
            with a single element (the tuple).
            Else converts the tuple to a list.

    # Returns
        A list.
    """
    if isinstance(x, list):
        return x
    if allow_tuple and isinstance(x, tuple):
        return list(x)
    return [x]


def rename(f, name):
    """rename functional.

    # Arguments
        f: A functional object.
        name: A unique (unused) string.

    # Returns
        updated f.
    """
    assert is_functional(f)
    assert str(name)
    f.layers[-1]._name = name
    return f



""" Functional class for SciANN.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.keras import backend as K
graph_unique_name = K.get_graph().unique_name

from tensorflow.python.keras.layers import Dense
from tensorflow.python.keras.layers import Activation
from tensorflow.python.keras.layers import Concatenate
from tensorflow.python.keras.layers import Lambda
from tensorflow.python.keras.models import Model
from tensorflow import tensordot, expand_dims

from ..utils import to_list, unpack_singleton, is_same_tensor, unique_tensors
from ..utils import default_weight_initializer
from ..utils import default_regularizer
from ..utils import validations, getitem
from ..utils import floatx, set_floatx
from ..utils import math
# from ..utils.transformers import FourierFeature
from ..utils.activations import SciActivation, get_activation
from ..utils import prepare_default_activations_and_initializers

from .field import Field


""" Configures the Functional object (Neural Network).

# Arguments
    fields: String or Field.
        [Sub-]Network outputs.
        It can be of type `String` - Associated fields will be created internally.
        It can be of type `Field` or `Functional`
    variables: Variable.
        [Sub-]Network inputs.
        It can be of type `Variable` or other Functional objects.
    hidden_layers: A list indicating neurons in the hidden layers.
        e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
    activation: defaulted to "tanh".
        Activation function for the hidden layers.
        Last layer will have a linear output.
    output_activation: defaulted to "linear".
        Activation function to be applied to the network output.
    res_net: (True, False). Constructs a resnet architecture.
        Defaulted to False.
    fourier_features: (True)
    kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
    bias_initializer: Initializer of the `Bias`, from `k.initializers`.
    kernel_regularizer: Regularizer for the kernel.
        To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
    bias_regularizer: Regularizer for the bias.
        To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
    dtype: data-type of the network parameters, can be
        ('float16', 'float32', 'float64').
        Note: Only network inputs should be set.
    trainable: Boolean.
        False if network is not trainable, True otherwise.
        Default value is True.

# Raises
    ValueError:
    TypeError:
"""
def SciFunctional(fields=None,
               variables=None,
               hidden_layers=None,
               activation="tanh",
               output_activation="linear",
               res_net=False,
               fourier=None,
               kernel_initializer=None,
               bias_initializer=None,
               kernel_regularizer=None,
               bias_regularizer=None,
               trainable=True,
               dtype='float64',
               **kwargs):
    # prepare hidden layers.
    if hidden_layers is None:
        hidden_layers = []
    else:
        hidden_layers = to_list(hidden_layers)
    # prepare kernel initializers.
    activations, def_biasinit, def_kerinit = \
        prepare_default_activations_and_initializers(
        len(hidden_layers) * [activation] + [output_activation]
    )
    if kernel_initializer is None:
        kernel_initializer = def_kerinit
    elif isinstance(kernel_initializer, (float, int)):
        kernel_initializer = default_weight_initializer(
            len(hidden_layers) * [activation] + [output_activation],
            'constant',
            scale=kernel_initializer
        )
    else:
        kernel_initializer = [kernel_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
    # prepare bias initializers.
    if bias_initializer is None:
        bias_initializer = def_biasinit
    elif isinstance(bias_initializer, (float, int)):
        bias_initializer = default_weight_initializer(
            len(hidden_layers) * [activation] + [output_activation],
            'constant',
            scale=bias_initializer
        )
    else:
        bias_initializer = [bias_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
    # prepare regularizers.
    kernel_regularizer = default_regularizer(kernel_regularizer)
    bias_regularizer = default_regularizer(bias_regularizer)
    # prepares fields.
    fields = to_list(fields)
    if all([isinstance(fld, str) for fld in fields]):
        output_fields = [
            Field(
                name=fld,
                dtype=dtype,
                kernel_initializer=kernel_initializer[-1],
                bias_initializer=bias_initializer[-1],
                kernel_regularizer=kernel_regularizer,
                bias_regularizer=bias_regularizer,
                trainable=trainable,
            )
            for fld in fields
        ]
    elif all([validations.is_field(fld) for fld in fields]):
        output_fields = fields
    else:
        raise TypeError(
            'Please provide a "list" of field names of'
            + ' type "String" or "Field" objects.'
        )
    # prepare inputs/outputs/layers.
    inputs = []
    layers = []
    variables = to_list(variables)
    if all([isinstance(var, Functional) for var in variables]):
        for var in variables:
            inputs += var.outputs
        # for var in variables:
        #     for lay in var.layers:
        #         layers.append(lay)
    else:
        raise TypeError(
            "Input error: Please provide a `list` of `Functional`s. \n"
            "Provided - {}".format(variables)
        )

    # Input layers.
    if len(inputs) == 1:
        net_input = inputs[0]
    else:
        layer = Concatenate(name=graph_unique_name('conct'))
        net_input = layer(inputs)

    # Define the output network.
    net = [net_input]

    # set up the Fourier layer.
    if fourier is not None:
        fourier = to_list(fourier)
        assert len(fourier) == len(inputs)
        fourier_output = []
        for x, w in zip(inputs, fourier):
            layers.append(
                FourierFeature(w, name=graph_unique_name("Fourier"))
            )
            fourier_output.append(layers[-1](x))
        if len(fourier_output)>1:
            layers.append(Concatenate(name=graph_unique_name('conct')))
            net[-1] = layers[-1](fourier_output)
        else:
            net[-1] = fourier_output[-1]

    # define the ResNet networks.
    if res_net is True:
        res_layers = []
        res_outputs = []
        for rl in ["U", "V", "H"]:
            layers.append(
                Dense(
                    hidden_layers[0],
                    kernel_initializer=kernel_initializer[0],
                    bias_initializer=bias_initializer[0],
                    kernel_regularizer=kernel_regularizer,
                    bias_regularizer=bias_regularizer,
                    trainable=trainable,
                    dtype=dtype,
                    name=graph_unique_name("DRes"+rl+"{:d}b".format(hidden_layers[0]))
                )
            )
            res_output = layers[-1](net_input)
            # Apply the activation.
            if activations[0].activation.__name__ != 'linear':
                layers.append(activations[0])
                res_outputs.append(layers[-1](res_output))
        net[-1] = res_outputs[-1]

    for nLay, nNeuron in enumerate(hidden_layers):
        # Add the layer.
        layer = Dense(
            nNeuron,
            kernel_initializer=kernel_initializer[nLay],
            bias_initializer=bias_initializer[nLay],
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            trainable=trainable,
            dtype=dtype,
            name=graph_unique_name("D{:d}b".format(nNeuron))
        )
        layers.append(layer)
        net[-1] = layer(net[-1])
        # Apply the activation.
        if activations[nLay].activation.__name__ != 'linear': #nLay<len(hidden_layers)-1 and
            layer = activations[nLay]
            layers.append(layer)
            net[-1] = layer(net[-1])
        # Add the resnet layer
        if res_net is True:
            layer = Lambda(lambda xs: (1-xs[0])*xs[1] + xs[0]*xs[2], name=graph_unique_name("ResLayer"))
            net[-1] = layer([net[-1]] + res_outputs[:2])

    # set up the Fourier layer.
    # if fourier is not None and len(fourier)>1:
    #     fourier_dims = [w if isinstance(w,int) else w.size for w in fourier]
    #     f_outputs = []
    #     for nNeuron in fourier_dims:
    #         layer = Dense(
    #             nNeuron,
    #             kernel_initializer=kernel_initializer[-1],
    #             bias_initializer=bias_initializer[-1],
    #             kernel_regularizer=kernel_regularizer,
    #             bias_regularizer=bias_regularizer,
    #             trainable=trainable,
    #             dtype=dtype,
    #             name=graph_unique_name("FourierD{:d}b".format(nNeuron))
    #         )
    #         layers.append(layer)
    #         net[-1] = layer(net[-1])
    #         # Apply the activation.
    #         if activations[-1].activation.__name__ != 'linear':  # nLay<len(hidden_layers)-1 and
    #             layer = activations[-1]
    #             layers.append(layer)
    #             net[-1] = layer(net[-1])
    #         f_outputs.append(net[-1])
    #     # Create the last output.
    #     layers.append(
    #         Lambda(lambda ys: tensordot(expand_dims(ys[0], -1), expand_dims(ys[1], -2), axes=[-1,-2]))
    #     )
    #     net[-1] = layer(f_outputs)

    # Assign to the output variable
    if len(net) == 1:
        net_output = net[0]
    else:
        raise ValueError("Legacy for Enrichment: Must be updated. ")
        layer = Concatenate(name=graph_unique_name('conct'))
        net_output = layer(net)

    # Define the final outputs of each network
    for out in output_fields:
        layers.append(out)
    outputs = []
    for out in output_fields:
        # add the activation on the output.
        if activations[-1].activation.__name__ != 'linear':
            layer = activations[-1]
            layers.append(layer)
            outputs.append(layer(out(net_output)))
        else:
            outputs.append(out(net_output))

    return Functional(
        inputs, outputs, layers
    )


#
#
# def Functional(object):
#     """ Configures the Functional object (Neural Network).
#
#     # Arguments
#         fields: String or Field.
#             [Sub-]Network outputs.
#             It can be of type `String` - Associated fields will be created internally.
#             It can be of type `Field` or `Functional`
#         variables: Variable.
#             [Sub-]Network inputs.
#             It can be of type `Variable` or other Functional objects.
#         hidden_layers: A list indicating neurons in the hidden layers.
#             e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
#         activation: defaulted to "tanh".
#             Activation function for the hidden layers.
#             Last layer will have a linear output.
#         output_activation: defaulted to "linear".
#             Activation function to be applied to the network output.
#         res_net: (True, False). Constructs a resnet architecture.
#             Defaulted to False.
#         fourier_features: (True)
#         kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
#         bias_initializer: Initializer of the `Bias`, from `k.initializers`.
#         kernel_regularizer: Regularizer for the kernel.
#             To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
#         bias_regularizer: Regularizer for the bias.
#             To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
#         dtype: data-type of the network parameters, can be
#             ('float16', 'float32', 'float64').
#             Note: Only network inputs should be set.
#         trainable: Boolean.
#             False if network is not trainable, True otherwise.
#             Default value is True.
#
#     # Raises
#         ValueError:
#         TypeError:
#     """
#     def __init__(self,
#                  fields=None,
#                  variables=None,
#                  hidden_layers=None,
#                  activation="tanh",
#                  output_activation="linear",
#                  res_net=False,
#                  fourier=None,
#                  kernel_initializer=None,
#                  bias_initializer=None,
#                  kernel_regularizer=None,
#                  bias_regularizer=None,
#                  dtype=None,
#                  trainable=True,
#                  **kwargs):
#         # check data-type.
#         if dtype is None:
#             dtype = K.floatx()
#         elif not K.floatx() == dtype:
#             K.set_floatx(dtype)
#         # prepare hidden layers.
#         if hidden_layers is None:
#             hidden_layers = []
#         else:
#             hidden_layers = to_list(hidden_layers)
#         # check for copy constructor.
#         if all([x in kwargs for x in ('inputs', 'outputs', 'layers')]):
#             self._inputs = kwargs['inputs'].copy()
#             self._outputs = kwargs['outputs'].copy()
#             self._layers = kwargs['layers'].copy()
#             self._set_model()
#             return
#         # prepare kernel initializers.
#         activations, def_biasinit, def_kerinit = \
#             prepare_default_activations_and_initializers(
#             len(hidden_layers) * [activation] + [output_activation]
#         )
#         if kernel_initializer is None:
#             kernel_initializer = def_kerinit
#         elif isinstance(kernel_initializer, (float, int)):
#             kernel_initializer = default_weight_initializer(
#                 len(hidden_layers) * [activation] + [output_activation],
#                 'constant',
#                 scale=kernel_initializer
#             )
#         else:
#             kernel_initializer = [kernel_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
#         # prepare bias initializers.
#         if bias_initializer is None:
#             bias_initializer = def_biasinit
#         elif isinstance(bias_initializer, (float, int)):
#             bias_initializer = default_weight_initializer(
#                 len(hidden_layers) * [activation] + [output_activation],
#                 'constant',
#                 scale=bias_initializer
#             )
#         else:
#             bias_initializer = [bias_initializer for l in len(hidden_layers) * [activation] + [output_activation]]
#         # prepare regularizers.
#         kernel_regularizer = default_regularizer(kernel_regularizer)
#         bias_regularizer = default_regularizer(bias_regularizer)
#         # prepares fields.
#         fields = to_list(fields)
#         if all([isinstance(fld, str) for fld in fields]):
#             output_fields = [
#                 Field(
#                     name=fld,
#                     dtype=dtype,
#                     kernel_initializer=kernel_initializer[-1],
#                     bias_initializer=bias_initializer[-1],
#                     kernel_regularizer=kernel_regularizer,
#                     bias_regularizer=bias_regularizer,
#                     trainable=trainable,
#                 )
#                 for fld in fields
#             ]
#         elif all([validations.is_field(fld) for fld in fields]):
#             output_fields = fields
#         else:
#             raise TypeError(
#                 'Please provide a "list" of field names of'
#                 + ' type "String" or "Field" objects.'
#             )
#         # prepare inputs/outputs/layers.
#         inputs = []
#         layers = []
#         variables = to_list(variables)
#         if all([isinstance(var, Functional) for var in variables]):
#             for var in variables:
#                 inputs += var.outputs
#             # for var in variables:
#             #     for lay in var.layers:
#             #         layers.append(lay)
#         else:
#             raise TypeError(
#                 "Input error: Please provide a `list` of `Functional`s. \n"
#                 "Provided - {}".format(variables)
#             )
#
#         # Input layers.
#         if len(inputs) == 1:
#             net_input = inputs[0]
#         else:
#             layer = Concatenate(name=graph_unique_name('conct'))
#             net_input = layer(inputs)
#
#         # Define the output network.
#         net = [net_input]
#
#         # set up the Fourier layer.
#         if fourier is not None:
#             fourier = to_list(fourier)
#             assert len(fourier) == len(inputs)
#             fourier_output = []
#             for x, w in zip(inputs, fourier):
#                 layers.append(
#                     FourierFeature(w, name=graph_unique_name("Fourier"))
#                 )
#                 fourier_output.append(layers[-1](x))
#             if len(fourier_output)>1:
#                 layers.append(Concatenate(name=graph_unique_name('conct')))
#                 net[-1] = layers[-1](fourier_output)
#             else:
#                 net[-1] = fourier_output[-1]
#
#         # define the ResNet networks.
#         if res_net is True:
#             res_layers = []
#             res_outputs = []
#             for rl in ["U", "V", "H"]:
#                 layers.append(
#                     Dense(
#                         hidden_layers[0],
#                         kernel_initializer=kernel_initializer[0],
#                         bias_initializer=bias_initializer[0],
#                         kernel_regularizer=kernel_regularizer,
#                         bias_regularizer=bias_regularizer,
#                         trainable=trainable,
#                         dtype=dtype,
#                         name=graph_unique_name("DRes"+rl+"{:d}b".format(hidden_layers[0]))
#                     )
#                 )
#                 res_output = layers[-1](net_input)
#                 # Apply the activation.
#                 if activations[0].activation.__name__ != 'linear':
#                     layers.append(activations[0])
#                     res_outputs.append(layers[-1](res_output))
#             net[-1] = res_outputs[-1]
#
#         for nLay, nNeuron in enumerate(hidden_layers):
#             # Add the layer.
#             layer = Dense(
#                 nNeuron,
#                 kernel_initializer=kernel_initializer[nLay],
#                 bias_initializer=bias_initializer[nLay],
#                 kernel_regularizer=kernel_regularizer,
#                 bias_regularizer=bias_regularizer,
#                 trainable=trainable,
#                 dtype=dtype,
#                 name=graph_unique_name("D{:d}b".format(nNeuron))
#             )
#             layers.append(layer)
#             net[-1] = layer(net[-1])
#             # Apply the activation.
#             if activations[nLay].activation.__name__ != 'linear': #nLay<len(hidden_layers)-1 and
#                 layer = activations[nLay]
#                 layers.append(layer)
#                 net[-1] = layer(net[-1])
#             # Add the resnet layer
#             if res_net is True:
#                 layer = Lambda(lambda xs: (1-xs[0])*xs[1] + xs[0]*xs[2], name=graph_unique_name("ResLayer"))
#                 net[-1] = layer([net[-1]] + res_outputs[:2])
#
#         # set up the Fourier layer.
#         # if fourier is not None and len(fourier)>1:
#         #     fourier_dims = [w if isinstance(w,int) else w.size for w in fourier]
#         #     f_outputs = []
#         #     for nNeuron in fourier_dims:
#         #         layer = Dense(
#         #             nNeuron,
#         #             kernel_initializer=kernel_initializer[-1],
#         #             bias_initializer=bias_initializer[-1],
#         #             kernel_regularizer=kernel_regularizer,
#         #             bias_regularizer=bias_regularizer,
#         #             trainable=trainable,
#         #             dtype=dtype,
#         #             name=graph_unique_name("FourierD{:d}b".format(nNeuron))
#         #         )
#         #         layers.append(layer)
#         #         net[-1] = layer(net[-1])
#         #         # Apply the activation.
#         #         if activations[-1].activation.__name__ != 'linear':  # nLay<len(hidden_layers)-1 and
#         #             layer = activations[-1]
#         #             layers.append(layer)
#         #             net[-1] = layer(net[-1])
#         #         f_outputs.append(net[-1])
#         #     # Create the last output.
#         #     layers.append(
#         #         Lambda(lambda ys: tensordot(expand_dims(ys[0], -1), expand_dims(ys[1], -2), axes=[-1,-2]))
#         #     )
#         #     net[-1] = layer(f_outputs)
#
#         # store output layers.
#         for out in output_fields:
#             layers.append(out)
#
#         # Assign to the output variable
#         if len(net) == 1:
#             net_output = net[0]
#         else:
#             raise ValueError("Legacy for Enrichment: Must be updated. ")
#             layer = Concatenate(name=graph_unique_name('conct'))
#             net_output = layer(net)
#
#         # Define the final outputs of each network
#         outputs = []
#         for out in output_fields:
#             # add the activation on the output.
#             if activations[-1].activation.__name__ != 'linear':
#                 layer = activations[-1]
#                 layers.append(layer)
#                 outputs.append(layer(out(net_output)))
#             else:
#                 outputs.append(out(net_output))
#
#         self._inputs = inputs
#         self._outputs = outputs
#         self._layers = layers
#         self._set_model()


# from tensorflow.python.util.tf_export import keras_export
# @keras_export('keras.model.SciFunctional')
class Functional(object):
    """ Configures the Functional object (Neural Network).

    # Arguments
        fields: String or Field.
            [Sub-]Network outputs.
            It can be of type `String` - Associated fields will be created internally.
            It can be of type `Field` or `Functional`
        variables: Variable.
            [Sub-]Network inputs.
            It can be of type `Variable` or other Functional objects.
        hidden_layers: A list indicating neurons in the hidden layers.
            e.g. [10, 100, 20] is a for hidden layers with 10, 100, 20, respectively.
        activation: defaulted to "tanh".
            Activation function for the hidden layers.
            Last layer will have a linear output.
        output_activation: defaulted to "linear".
            Activation function to be applied to the network output.
        res_net: (True, False). Constructs a resnet architecture.
            Defaulted to False.
        fourier_features: (True)
        kernel_initializer: Initializer of the `Kernel`, from `k.initializers`.
        bias_initializer: Initializer of the `Bias`, from `k.initializers`.
        kernel_regularizer: Regularizer for the kernel.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        bias_regularizer: Regularizer for the bias.
            To set l1 and l2 to custom values, pass [l1, l2] or {'l1':l1, 'l2':l2}.
        dtype: data-type of the network parameters, can be
            ('float16', 'float32', 'float64').
            Note: Only network inputs should be set.
        trainable: Boolean.
            False if network is not trainable, True otherwise.
            Default value is True.

    # Raises
        ValueError:
        TypeError:
    """
    def __init__(self, inputs, outputs, layers):
        # check data-type.
        self.inputs = inputs.copy()
        self.outputs = outputs.copy()
        self.layers = layers.copy()
        self._set_model()

    def _set_model(self, inputs, outputs, layers):
        self.inputs = inputs
        self.outputs = outputs
        self.layers = layers

    def eval(self, *kwargs):
        """ Evaluates the functional object for a given input.

        # Arguments
            (SciModel, Xs): 
                Evalutes the functional object from the beginning 
                    of the graph defined with SciModel. 
                    The Xs should match those of SciModel. 
            
            (Xs):
                Evaluates the functional object from inputs of the functional. 
                    Xs should match those of inputs to the functional. 
                    
        # Returns
            Numpy array of dimensions of network outputs. 

        # Raises
            ValueError:
            TypeError:
        """
        if len(kwargs) == 1:
            model = self.model
            # read data.
            mesh = kwargs[0]
        elif len(kwargs) == 2:
            if validations.is_scimodel(kwargs[0]):
                model = K.function(kwargs[0].model.inputs, self.outputs)
            else:
                raise ValueError(
                    'Expected a SciModel object for the first arg. '
                )
            mesh = kwargs[1]
        else:
            raise NotImplemented()
        x_pred = to_list(mesh.copy())
        # To have unified output for postprocessing - limitted support.
        shape_default = x_pred[0].shape if all([x.shape==x_pred[0].shape for x in x_pred]) else None
        # prepare X,Y data.
        for i, (x, xt) in enumerate(zip(x_pred, model.inputs)):
            x_shape = tuple(xt.get_shape().as_list())
            if x.shape != x_shape:
                try:
                    x_pred[i] = x.reshape((-1,) + x_shape[1:])
                except:
                    print(
                        'Could not automatically convert the inputs to be ' 
                        'of the same size as the expected input tensors. ' 
                        'Please provide inputs of the same dimension as the `Variables`. '
                    )
                    assert False

        y_pred = to_list(model(x_pred))

        if shape_default is not None:
            try:
                y_pred = [y.reshape(shape_default) for y in y_pred]
            except:
                print("Input and output dimensions need re-adjustment for post-processing.")

        return unpack_singleton(y_pred)


    @property
    def layers(self):
        return self._layers

    @layers.setter
    def layers(self, value):
        self._layers = value

    @property
    def inputs(self):
        return self._inputs

    @inputs.setter
    def inputs(self, value):
        self._inputs = value

    @property
    def outputs(self):
        return self._outputs

    @outputs.setter
    def outputs(self, value):
        self._outputs = value

    @property
    def model(self):
        self._set_model()
        return self._model

    @property
    def name(self):
        return self._layers[-1].name
    
    def _set_model(self):
        if hasattr(self, '_model'):
            if is_same_tensor(self._inputs, self._model.inputs) and \
               is_same_tensor(self._outputs, self._model.outputs):
               return
        self._model = K.function(
            unique_tensors(self._inputs),
            self._outputs
        )

    def get_weights(self, at_layer=None):
        """ Get the weights and biases of different layers.

        # Arguments
            at_layer: 
                Get the weights of a specific layer. 
            
        # Returns
            List of numpy array. 
        """
        return [l.get_weights() for l in self.layers]

    def set_weights(self, weights):
        """ Set the weights and biases of different layers.

        # Arguments
            weights: Should take the dimensions as the output of ".get_weights"
            
        # Returns 
        """
        try:
            for l, w in zip(self.layers, weights):
                l.set_weights(w)
        except:
            raise ValueError(
                'Provide data exactly the same as .get_weights() outputs. '
            )

    def count_params(self):
        """ Total number of parameters of a functional.

        # Arguments
            
        # Returns 
            Total number of parameters.
        """
        return sum([l.count_params() for l in self.layers])

    def copy(self):
        return Functional(
            inputs=self.inputs,
            outputs=self.outputs,
            layers=self.layers
        )

    def append_to_layers(self, layers):
        if self.layers is not layers:
            cl = [x.name for x in self.layers]
            for x in layers:
                if not x.name in cl:
                    self.layers += [x]

    def append_to_inputs(self, inputs):
        if self.inputs is not inputs:
            cl = [x.name for x in self.inputs]
            for x in inputs:
                if not x.name in cl:
                    self.inputs.append(x)

    def append_to_outputs(self, outputs):
        self._outputs += to_list(outputs)

    def set_trainable(self, val, layers=None):
        """ Set the weights and biases of a functional object trainable or not-trainable.
        Note: The SciModel should be called after this.

        # Arguments
            val: (Ture, False)
            layers: list of layers to be set trainable or non-trainable.
                defaulted to None.
            
        # Returns 
        """
        print("Warning: Call 'SciModel' after using set_trainable")
        if isinstance(val, bool):
            if layers is None:
                for l in self._layers:
                    l.trainable = val
            else:
                for li in to_list(layers):
                    i = -1
                    for l in self._layers:
                        if l.count_params() > 0:
                            i += 1
                        if li == i:
                            l.trainable = val
                            break
        else:
            raise ValueError('Expected a boolean value: True or False')

    def reinitialize_weights(self):
        """ Re-initialize the weights and biases of a functional object.

        # Arguments

        # Returns 
        
        """
        for lay in self.layers:
            if hasattr(lay, 'kernel_initializer') and lay.kernel is not None:
                K.set_value(lay.kernel, lay.kernel_initializer(lay.kernel.shape))
            if hasattr(lay, 'bias_initializer') and lay.bias is not None:
                K.set_value(lay.bias, lay.bias_initializer(lay.bias.shape))

    def split(self):
        """ In the case of `Functional` with multiple outputs,
            you can split the outputs and get an associated functional.

        # Returns
            (f1, f2, ...): Tuple of splitted `Functional` objects
                associated to each output.
        """
        if len(self._outputs)==1:
            return self
        fs = ()
        # total number of outputs to get splitted.
        nr = len(self._outputs)
        # associated to each output, there is a layer to be splitted.
        lays = self.layers[:-nr]
        for out, lay in zip(self._outputs, self._layers[-nr:]):
            # copy constructor for functional.
            f = Functional(
                inputs = to_list(self.inputs),
                outputs = to_list(out),
                layers = lays + to_list(lay)
            )
            fs += (f,)
        return fs

    def __call__(self):
        return self.outputs

    def __pos__(self):
        return self

    def __neg__(self):
        return self*-1.0

    def __add__(self, other):
        return math.add(self, other)

    def __radd__(self, other):
        return math.radd(self, other)

    def __sub__(self, other):
        return math.sub(self, other)

    def __rsub__(self, other):
        return math.rsub(self, other)

    def __mul__(self, other):
        return math.mul(self, other)

    def __rmul__(self, other):
        return math.rmul(self, other)

    def __truediv__(self, other):
        return math.div(self, other)

    def __rtruediv__(self, other):
        return math.rdiv(self, other)

    def __pow__(self, power):
        return math.pow(self, power)

    def __getitem__(self, item):
        return getitem(self, item)

    def diff(self, *args, **kwargs):
        return math.diff(self, *args, **kwargs)

    def __eq__(self, other):
        return math.equal(self, other)

    def __ne__(self, other):
        return math.not_equal(self, other)

    def __gt__(self, other):
        return math.greater(self, other)

    def __ge__(self, other):
        return math.greater_equal(self, other)

    def __lt__(self, other):
        return math.less(self, other)

    def __le__(self, other):
        return math.less_equal(self, other)

    @classmethod
    def get_class(cls):
        return Functional
